
import json
import time
import datetime
import re

import tqdm
import pandas as pd

from hashlib import md5
from typing import Iterable
from functools import reduce
from typing import Any
from typing import List

from websocket import create_connection
from termcolor import colored

class SmallModel:
    remote_host_ :str = None
    def __init__(self,name, remote=None):
        if remote is None:
            remote = self.__class__.remote_host_
        assert remote is not None
        self.remote_host  = remote
        self.name = name
        self._ok = False
        self.purpose = "unKnow"
    
    
    def status(self):
        try:
            ws = create_connection(f"ws://{self.remote_host}:15000")
            user_id = md5(time.asctime().encode()).hexdigest()
            TODAY = datetime.datetime.now()
            PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
            ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
            res = ws.recv()
            if res != "ok":
                print(colored("[info]:","yellow") ,res)
                raise Exception("password error")
            res = self.send_and_recv(json.dumps({"embed_documents":[self.name], "method":"status"}),ws)
            return res
        except Exception as e:
            raise e
        finally:
            ws.close()
    
    @classmethod
    def from_remote(cls, name,remote=None):
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        model = cls(name, remote)
        model.try_load_in_remote()
        return model
    
    @classmethod
    def gpu_info_remote(cls, remote=None):
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        ws = create_connection(f"ws://{remote}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        res = cls.send_and_recv(json.dumps({"embed_documents":["all"], "method":"gpu"}),ws)["embed"]
        ress = []
        for o in res:
            used, allcan = o.split("/",1)
            ress.append({"used":int(used.strip()[:-3]), "all":int(allcan.strip()[:-3]), "c":"MiB"})
        return ress
    
    @classmethod
    def clean_all(cls, remote=None):
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        all_models = cls.show_all_loaded_models(remote)
        ws = create_connection(f"ws://{remote}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        res = cls.send_and_recv(json.dumps({"embed_documents":all_models, "method":"clean"}),ws)
        return res
    
    @classmethod
    def show_all_loaded_models(cls, remote=None):
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        ws = create_connection(f"ws://{remote}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        res = cls.send_and_recv(json.dumps({"embed_documents":["all"], "method":"ls"}),ws)
        return res["embed"]

    @classmethod
    def show_all_models(cls, remote=None):
        if remote is None:
            remote = cls.remote_host_
        else:
            cls.remote_host_ = remote
        assert remote is not None
        ws = create_connection(f"ws://{remote}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        res = cls.send_and_recv(json.dumps({"embed_documents":["all"], "method":"ls-all"}),ws)
        return res["embed"]
        
    def clean(self):
        try:
            ws = create_connection(f"ws://{self.remote_host}:15000")
            user_id = md5(time.asctime().encode()).hexdigest()
            TODAY = datetime.datetime.now()
            PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
            ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
            res = ws.recv()
            if res != "ok":
                print(colored("[info]:","yellow") ,res)
                raise Exception("password error")
            res = self.send_and_recv(json.dumps({"embed_documents":[self.name], "method":"clean"}),ws)
            return res["embed"]
        except Exception as e:
            raise e
        finally:
            ws.close()

    
    def down_remote(self, try_time=3):
        try:
            ws = create_connection(f"ws://{self.remote_host}:15000")
            user_id = md5(time.asctime().encode()).hexdigest()
            TODAY = datetime.datetime.now()
            PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
            ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
            res = ws.recv()
            if res != "ok":
                print(colored("[info]:","yellow") ,res)
                raise Exception("password error")
            res = self.send_and_recv(json.dumps({"embed_documents":[self.name], "method":"clone"}),ws)
            time.sleep(2)
            res = self.msg()
            if "git clone failed: exit status 128" in res:
                print(colored("[Err : ]:","yellow") , colored(res,"red"))
                if try_time > 0:
                    return self.down_remote(try_time-1)
            return res["embed"]
            
        except Exception as e:
            raise e
        finally:
            ws.close()

    @classmethod
    def send_and_recv(cls, data, ws):
        try:
            T = len(data)// (1024*102)
            bart = tqdm.tqdm(total=T,desc=colored(" + sending data","cyan"))
            bart.leave = False
            for i in range(0, len(data), 1024*102):
                bart.update(1)
                ws.send(data[i:i+1024*102])
            bart.clear()
            bart.close()

            ws.send("[STOP]")
            message = ""
            total = int(ws.recv())
            bar = tqdm.tqdm(desc=colored(" + receiving data","cyan", attrs=["bold"]), total=total)
            bar.leave = False
            while 1:
                res = ws.recv()
                message += res
                bar.update(len(res))
                if message.endswith("[STOP]"):
                    message = message[:-6]
                    break
            bar.clear()
            bar.close()
            msg = json.loads(message)
            return msg
        except Exception as e:
            raise e
    
    def msg(self):
        try:
            ws = create_connection(f"ws://{self.remote_host}:15000")
            user_id = md5(time.asctime().encode()).hexdigest()
            TODAY = datetime.datetime.now()
            PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
            ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
            res = ws.recv()
            if res != "ok":
                print(colored("[info]:","yellow") ,res)
                raise Exception("password error")
            res = self.send_and_recv(json.dumps({"embed_documents":[self.name], "method":"msg"}),ws)
            return res.get("embed","")
        except Exception as e:
            raise e
        finally:
            ws.close()
    
    def change_remote_name(self, new_name):
        assert "/" not in new_name
        assert " " not in new_name
        ss = re.findall(r"[\w\-\_]+",new_name)
        assert len(ss) == 1 and ss[0] == new_name
        if self.name in self.show_all_models():
            try:
                ws = create_connection(f"ws://{self.remote_host}:15000")
                user_id = md5(time.asctime().encode()).hexdigest()
                TODAY = datetime.datetime.now()
                PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
                ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
                res = ws.recv()
                if res != "ok":
                    print(colored("[info]:","yellow") ,res)
                    raise Exception("password error")
                res = self.send_and_recv(json.dumps({"embed_documents":[self.name, new_name], "method":"change_name"}),ws)
                return res["embed"]
            except Exception as e:
                raise e
            finally:
                ws.close()
        else:
            raise Exception("model not exists")

    def check(self):
        self._ok = False
        name = self.name
        if "/" in name:
            name = name.rsplit("/",1)[-1]

        ws = create_connection(f"ws://{self.remote_host}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        reply =  self.send_and_recv(json.dumps({"embed_documents":[name], "method":"check"}), ws)["embed"]
        if reply.startswith("ok"):
            if ":" in reply:
                self.purpose = reply.split(":",1)[-1].strip()
            self._ok = True
        else:
            self.try_load_in_remote()

        return self._ok
    
    def try_load_in_remote(self):
        try:
            self._ok = False
            name = self.name
            if "/" in name:
                name = name.rsplit("/",1)[-1]
                
            ws = create_connection(f"ws://{self.remote_host}:15000")
            user_id = md5(time.asctime().encode()).hexdigest()
            TODAY = datetime.datetime.now()
            PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
            ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
            res = ws.recv()
            if res != "ok":
                print(colored("[info]:","yellow") ,res)
                raise Exception("password error")
            reply = self.send_and_recv(json.dumps({"embed_documents":[name], "method":"load"}), ws)["embed"]
            if reply.startswith("ok"):
                self._ok = True
                if ":" in reply:
                    self.purpose = reply.split(":",1)[-1].strip()
                return True
                
            return False
        except Exception as e:
            raise e
        finally:
            ws.close()

    # def show_remote_models(self):
    #     try:
    #         name = self.name
    #         if "/" in name:
    #             name = name.rsplit("/",1)[-1]
                
    #         ws = create_connection(f"ws://{self.remote_host}:15000")
    #         user_id = md5(time.asctime().encode()).hexdigest()
    #         TODAY = datetime.datetime.now()
    #         PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
    #         ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
    #         res = ws.recv()
    #         if res != "ok":
    #             print(colored("[info]:","yellow") ,res)
    #             raise Exception("password error")
    #         return self.send_and_recv(json.dumps({"embed_documents":[name], "method":"show"}), ws)["embed"]
    #     except Exception as e:
    #         raise e
    #     finally:
    #         ws.close()

    def __call__(self, args: List[str], pandas=True) -> Any:
        if isinstance(args, str):
            args = [args]
        assert isinstance(args, (list, tuple,Iterable,))
        if not self._ok:
            self.check()

        if not self._ok :
            raise Exception("remote's service no such model deployed"+self.name)
        ws = create_connection(f"ws://{self.remote_host}:15000")
        user_id = md5(time.asctime().encode()).hexdigest()
        TODAY = datetime.datetime.now()
        PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
        ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
        # time.sleep(0.5)
        res = ws.recv()
        if res != "ok":
            print(colored("[info]:","yellow") ,res)
            raise Exception("password error")
        name = self.name
        if "/" in name:
            name = name.rsplit("/",1)[-1]
            
        data = json.dumps({"embed_documents":args, "method": name})
        try:
            msg = self.send_and_recv(data, ws)
            o = self._merged(msg["embed"])
            if pandas:
                oss = []
                if len(o) > 0 and isinstance(o[0], str):
                    return o
                for i,ii in enumerate(o):
                    ar = args[i]
                    if isinstance(ii, dict ):
                        ii["input"] = ar
                    oss.append(ii)
                if isinstance(oss[0], dict):
                    return pd.DataFrame(oss)
                return oss
            return o
        except Exception as e:
            raise e
            # import ipdb;ipdb.set_trace()
    
    def _merged(self, res_list):
        res = []
        for one_obj in res_list:
            if isinstance(one_obj, list) :
                if len(one_obj) > 1:
                    if isinstance(one_obj[0], dict):
                        res.append({k:reduce(lambda x,y: x+y , map(lambda x: x[k] ,one_obj)) for k in one_obj[0]})
                    else:
                        res.append(one_obj)
                else:
                    res.append(one_obj[0])
            else:
                res.append(one_obj)
        return res


import os
from pathlib import Path
from multiprocessing import Process
from subprocess import Popen, PIPE, STDOUT
import tempfile
class Git:
    
    def __init__(self, querys):
        
        self.querys = querys
    
    def start(self):
        print("start download:"+str(self.querys))
        p = Process(target=self.down, args=(self.querys,))
        p.start()
        # p.join()

    

    def down(self, querys):
        for query in querys:
            
            msg = self.download(query)
            print(msg)
    
    def download(self, name):
        try:
            assert name.startswith("https://huggingface")
            home = str((Path("~") / ".cache" ).expanduser())
            if "/" in name:
                p = Path(home) / name.rsplit("/")[-1]
                if p.exists():
                    os.popen("rm -rf "+str(p))
                
            cmd  = f"cd {home} && git lfs clone  {name} "
            f = Path(tempfile.tempdir) / "down.log"
            with open(f, "w") as f:
                proc = Popen(cmd, shell=True, env=os.environ, stdout=f, stderr=f, close_fds=True)
                # proc.stdout.read()
                proc.wait()
                
                # os.spawnl(os.P_NOWAIT, cmd)

                
            return "ok"
        except AssertionError:
            return name+ " : error in repo must start with https://huggingface"
        except Exception as e:
            print(e)
            return str(e)
            


from sklearn.cluster import KMeans, DBSCAN

from sklearn.metrics.pairwise import cosine_similarity, euclidean_distances
from websocket import create_connection
import datetime
import time
from hashlib import md5
import json
from termcolor import colored
import tqdm
import termcolor


def cluster(data, n_clusters=2,eps=0.5,min_samples=3, method='kmeans'):
    
    if method == 'kmeans':
        kmeans = KMeans(n_clusters=n_clusters, random_state=0, n_init='auto').fit(data)
        return kmeans.labels_
    elif method == 'dbscan':
        db = DBSCAN(metric='euclidean',eps=eps, min_samples=min_samples).fit(data)
        return db.labels_
    elif method == 'cosine':
        sim_vec = cosine_similarity(data).tolist()
    else:
        raise ValueError('method must be kmeans or dbscan')


def similarity(data, method='cosine'):
    if method == 'cosine':
        return cosine_similarity(data).tolist()
    elif method == 'euclidean':
        return euclidean_distances(data).tolist()
    else:
        raise ValueError('method must be cosine or euclidean')


def papers(remote_host,texts, method="dbscan", eps=0.3, min_samples=3, n_clusters=2,filter_noise=None):
    """
    @method : str, default='dbscan' , can used 'kmeans' or 'dbscan' or 'cosine' or 'euclidean'
        @eps : float, default=0.3, used in dbscan
        @min_samples : int, default=3, used in dbscan
        @n_clusters : int, default=2, used in kmeans
    
    """
    ws = create_connection(f"ws://{remote_host}:15000")
    user_id = md5(time.asctime().encode()).hexdigest()
    TODAY = datetime.datetime.now()
    PASSWORD = "ADSFADSGADSHDAFHDSG@#%!@#T%DSAGADSHDFAGSY@#%@!#^%@#$Y^#$TYDGVDFSGDS!@$!@$" + f"{TODAY.year}-{TODAY.month}"
    ws.send(json.dumps({"user_id":user_id, "password":PASSWORD}))
    # time.sleep(0.5)
    res = ws.recv()
    if res != "ok":
        print(colored("[info]:","yellow") ,res)
        raise Exception("password error")
    
    data = json.dumps({
        "embed_documents":texts,
        "method":method,
        "n_clusters":n_clusters,
        "filter_noise":filter_noise,
        "eps":eps,
        "min_samples":min_samples

    })
    msg = send_and_recv(data, ws)
    return msg


def send_and_recv(data, ws):
    try:
        T = len(data)// 1024*102
        bart = tqdm.tqdm(total=T,desc=termcolor.colored(" + sending data","cyan"))
        bart.leave = False
        for i in range(0, len(data), 1024*102):
            bart.update(1)
            ws.send(data[i:i+1024*102])
        bart.clear()
        bart.close()
        ws.send("[STOP]")
        message = ""
        total = int(ws.recv())
        bar = tqdm.tqdm(desc=termcolor.colored(" + receiving data","cyan", attrs=["bold"]), total=total)
        bar.leave = False
        while 1:
            res = ws.recv()
            message += res
            bar.update(len(res))
            if message.endswith("[STOP]"):
                message = message[:-6]
                break
        
        bar.clear()
        bar.close()
        msg = json.loads(message)
        return msg
    except Exception as e:
        raise e



from setuptools import setup, find_packages


setup(name='chatglm-llm',
    version='1.5.4',
    description='chatglm llm',
    url='https://github.com/xxx',
    author='auth',
    author_email='xxx@gmail.com',
    license='MIT',
    include_package_data=True,
    zip_safe=False,
    packages=find_packages(),
    extras_require={
        "all": [
            'sentence_transformers',
            'tensorboard',
            "protobuf",
            "fschat==0.2.2",
            "cpm_kernels",
            "mdtex2html",
            "sentencepiece",
            "accelerate",
            "scikit-learn",
            'torch',    
            'termcolor',
            'transformers',
        ],
    },
    install_requires=[
        'requests',
        'termcolor',
        'tqdm',
        'gptcache',
        'numpy',
        'pypdf',
        "scikit-learn",
        'langchain',
        'websockets',
        'websocket-client',
        'unstructured',
        'aiowebsocket',
        ],
    entry_points={
        'console_scripts': [
            'chatglm-web=chatglm_src.cmd:main',
        ]
    },

)
